<template>
  <div :class="['my-tree',treeStyle]">
    <div
      class="serch-box"
      v-if="openSerch"
    >
      <div class="serch">
        <input
          type="textarea"
          placeholder="请输入关键字搜索"
          v-model.trim="serchVal"
          @keyup="serch(serchVal)"
        />
        <img
          src="./img/close-icon.png"
          class="clear-icon"
          @click="reset"
        >
      </div>
      <div
        class="text"
        @click="serch(serchVal)"
      >搜索
      </div>
    </div>
    <div
      class="my-tree-row"
      v-for="(item) in treeList"
      :key="item.id"
      :style="{
        paddingLeft: `${item.rank * indent}px`,
        height: item.isShow ? item.height : '0px',
        pointerEvents: item.isShow ? 'auto' : 'none',
      }"
      :class="[
        item.isShow ? 'isShow' : '',
      ]"
      @click="rowClick(item)"
    >
      <!-- 默认作用域插槽 -->
      <slot :row="item">
        <div :class="['my-tree-content',params.border ? 'border' : '']">
          <div class="prefix-icon">
            <!-- 前置作用域插槽 -->
            <slot
              name="prefix"
              :row="item"
            >
              <img :src="require(`${params.prefixIcon}`)">
            </slot>
          </div>
          <div class="middle-conetnt">
            <!-- 中置作用域插槽  -->
            <slot
              name="content"
              :row="item"
            >
              <span v-html="brightenKeyword(item.name,serchVal)"></span>
            </slot>
          </div>

          <div class="middle-last">
            <div class="post-icon">
              <!-- 后置作用域插槽 -->
              <slot
                name="post"
                :row="item"
              >
                <img
                  v-if="item.disabled"
                  :src="require(`${params.disabledIcon}`)"
                  style="width:17px"
                >
                <img
                  v-else
                  :class="[item.showChild ? 'arrowTransform' :'arrowTransformReturn']"
                  :src="item.lastRank ? '' :require(`${params.postIcon}`)"
                >
              </slot>
            </div>
          </div>
        </div>
      </slot>

    </div>
  </div>
</template>

<script>
export default {
  name: 'MyTree',
  props: {
    tree: { // 树形数据源
      type: Array,
      default: () => []
    },
    sonOptions: { // 多项子集配置
      type: Array,
      default: () => [{
        childrenkey: 'children',
        nameKey: 'name',
        idKey: 'id',
        disabledKey: 'disabled',
        height: '32px'
      }]
    },
    topChild: { // 顶层对象含义
      type: String,
      default: 'children'
    },
    treeParams: { // 个性化样式配置
      type: Object,
      default: () => {
      }
    },
    indent: {   //  相邻级节点间的水平缩进，单位为像素
      type: Number,
      default: 18
    },
    defaultExpandAll: {  // 是否默认展开所有节点
      type: Boolean,
      default: false
    },
    foldAll: {  // 折叠时关闭所有已经打开的子集，再次打开时需要一级一级打开
      type: Boolean,
      default: false
    },
    openSerch: {    // 开启搜索
      type: Boolean,
      default: false
    },
    treeStyle: {  // 自定义样式类名
      type: String,
      default: ''
    }
  },
  data() {
    return {
      treeList: [],
      initialData: [],
      params: {
        prefixIcon: './img/prefix-icon.png',  // 前置图标
        postIcon: './img/post-icon.png',  // 后置图标（自动旋转）
        disabledIcon: './img/disabled-icon.png', // 禁用图标
        border: false
      },
      serchVal: ''
    }
  },
  watch: {
    tree: {
      handler(tree) {
        this.params = Object.assign(this.params, this.treeParams)
        // console.time('数据处理时间：')
        // 异步执行避免后续功能出错
        setTimeout(() => {
          this.initialData = this.flattenTree(tree)
          this.treeList = this.deepClone(this.initialData)
          // console.timeEnd('数据处理时间：')
          this.$emit('is-finished', this.initialData)
        }, 0)
      },
      immediate: true // 初始化执行
    }
  },
  mounted() {
  },
  methods: {
    // 扁平化树形结构
    flattenTree(tree = [], rank = 0, parentIds = []) {
      const {sonOptions, topChild, defaultExpandAll, flattenTree} = this

      return tree.reduce((concatArr, item) => {
        const {
          nameKey,
          idKey,
          disabledKey,
          height
        } = sonOptions.find(i => {
          return i.childrenkey === (item.childrenkey || topChild)
        })
        let newItem = {
          ...item,
          name: item[nameKey] || item.name, // 显示文字
          id: item[idKey] || item.id, // 唯一标识符
          height: height || '32px', // 高度
          isShow: defaultExpandAll || rank === 0, // 自身是否显示
          showChild: (defaultExpandAll && !item.lastRank) || false, //子级是否展开
          childrenkey: item.childrenkey || topChild, // 父级下层级代表
          rank, // 层级
          lastRank: false, // 是否为最底层
          disabled: !!(item[disabledKey] || item.disabled), // 是否禁用
          parentIds, // 父级id数组
          parentId: parentIds[parentIds.length - 1] || 0 // 父级id
        }
        // 合并所有类型的子集数组并打上类型标记
        let children = sonOptions.map(({childrenkey}) => {
          delete newItem[childrenkey] // 删除扁平化数组对象中的所有子集数组
          return childrenkey
        }).map(childrenkey => {
          if (Array.isArray(item[childrenkey]) && item[childrenkey].length) {
            return item[childrenkey].map(_item => {
              return {childrenkey, ..._item}
            })
          }
        }).filter(item => item).reduce((sonConcat, son) => {
          if (Array.isArray(son) && son.length) return sonConcat.concat(son)
        }, [])
        // 判断是否到达底层
        if (Array.isArray(children) && children.length) {
          let parents = [...parentIds]
          parents.push(item[sonOptions.find(i => i.childrenkey === topChild).idKey])
          return concatArr.concat(newItem, flattenTree(children, rank + 1, parents))
        } else {
          newItem.lastRank = true
          return concatArr.concat(newItem)
        }
      }, [])
    },

    //使用递归的方式实现数组、对象的深拷贝
    deepClone(obj) {
      // 只拷贝对象
      if (typeof obj !== 'object') return
      // 根据obj的类型判断是新建一个数组还是一个对象
      let newObj = obj instanceof Array ? [] : {}
      for (let key in obj) {
        // 遍历obj,并且判断是obj的属性才拷贝
        if (obj.hasOwnProperty(key)) {
          // 判断属性值的类型，如果是对象递归调用深拷贝
          newObj[key] = typeof obj[key] === 'object' ? this.deepClone(obj[key]) : obj[key]
        }
      }
      return newObj
    },

    // 行点击
    rowClick(row) {
      const {lastRank, id, disabled} = row
      const {foldAll} = this
      if (disabled) return this.$emit('disabled-item-click', row)
      if (lastRank) return this.$emit('tree-item-click', row)
      row.showChild = !row.showChild
      this.treeList.forEach((item, index, arr) => {
        if (!row.showChild) {
          // 关闭子集
          if (item.parentIds.includes(id)) {
            item.isShow = false
            // 是否开启关闭折叠所有子集项功能
            if (foldAll) {
              arr.forEach(({id}) => {
                if (item.parentId === id) item.showChild = false
              })
            }
          }
        } else {
          // 打开子集
          const clickItem = arr.find(({id}) => id === item.parentId) || {}
          if (item.parentIds.includes(id) && clickItem.showChild) {
            item.isShow = true
          }
        }
      })
    },

    // 展开所有列表
    expandAll() {
      this.treeList.forEach((item, index, arr) => {
        if (!item.lastRank && !item.disabled) item.showChild = true
        item.isShow = true
        // 将禁用项的组件的子集关闭
        arr.forEach(_item => {
          if (item.parentIds.includes(_item.id) && _item.disabled) {
            item.isShow = false
            item.showChild = false
          }
        })
      })
    },

    // 搜索
    serch(serchVal) {
      this.serchVal.trim()
      const {initialData, tree, flattenTree, expandAll} = this
      if (!serchVal) {
        return this.treeList = flattenTree(tree)
      }
      const treeList = initialData.filter(({id}) => {
        return [...new Set(initialData.filter(({name}) => {
          return name.includes(serchVal) || serchVal.includes(name)
        }).reduce((concatArr, {parentIds, id, name}) => {
          if (Array.isArray(parentIds) && parentIds.length) return concatArr.concat(parentIds, id)
        }, []))].includes(id)
      })
      let max = 0
      treeList.map(({rank}) => rank).forEach(item => max = item > max ? item : max)
      treeList.forEach(item => {
        if (item.rank === max) {
          item.showChild = false
          item.lastRank = true
        }
      })
      this.treeList = treeList
      expandAll()
    },

    // 重置搜索内容
    reset() {
      this.serchVal = ''
      this.treeList = this.flattenTree(this.tree)
    },

    // 高亮搜索内容
    brightenKeyword(val, keyword) {
      if (keyword.length) {
        let keywordArr = keyword.split('')
        val = val + ''
        keywordArr.forEach(item => {
          if (val.indexOf(item) !== -1 && item !== ' ') {
            val = val.replace(
              new RegExp(item, 'g'),
              '<font color="#f75353">' + item + '</font>'
            )
          }
        })
      }
      return val
    }
  }
}
</script>

<style scoped>
@import "./style/myTree.css";
</style>
